<?php

namespace Drupal\config_enforce_devel;

use Drupal\Component\Utility\Crypt;
use Drupal\Core\File\FileSystemInterface;
use Drupal\config_enforce\EnforcedConfig as ReadOnlyEnforcedConfig;

/**
 * Enforces configuration.
 */
class EnforcedConfig extends ReadOnlyEnforcedConfig {

  /**
   * Provide Config Devel module with configs to read from disk.
   *
   * @see hook_config_devel_auto_import()
   */
  public static function autoImport() {
    $auto_import = [];
    foreach (self::getEnforcedConfigs() as $config_name => $setting) {
      $auto_import[$config_name] = [
        'filename' => $setting['config_file_path'],
        'hash' => isset($setting['hash']) ? $setting['hash'] : '',
      ];
    }
    return $auto_import;
  }

  /**
   * Provide Config Devel module with config files to write to disk.
   *
   * @see hook_config_devel_auto_export()
   */
  public static function autoExport() {
    $auto_export = [];
    foreach (self::getEnforcedConfigs() as $config_name => $settings) {
      $auto_export[] = $settings['config_file_path'];
    }
    return $auto_export;
  }

  /**
   * Save hashes of auto-imported config files.
   *
   * @see config_enforce_devel_config_devel_post_auto_import()
   */
  public static function postAutoImport(array $auto_import) {
    $enforced_configs = self::getEnforcedConfigs();
    foreach ($auto_import as $config_name => $info) {
      $collection = new EnforcedConfigRegistry($enforced_configs[$config_name]['target_module']);
      if ($config_name == $collection->getConfigName()) {
	// The current config is an enforced_configs registry. As such, its
	// contents will change as soon as it's saved below, and so will the
	// file hash. As a result, we set the hash explicitly, to avoid
	// spurious changes.
	$info['hash'] = 'intentionally_invalid';
      }
      if (empty($info['hash'])) {
        $contents = @file_get_contents($info['filename']);
        $info['hash'] = Crypt::hashBase64($contents);
      }
      $enforced_configs[$config_name]['hash'] = $info['hash'];
      $collection->getEditableConfig()
        ->set(self::encode($config_name), $enforced_configs[$config_name])
        ->save();
    }
  }

  /**
   * Remove a config object from the enforced configs.
   */
  public function clear() {
    (new EnforcedConfigRegistry($this->targetModule))->getEditableConfig()
      ->clear(self::encode($this->configName))
      ->save();
    return $this;
  }

  /**
   * Add or update a config object in the enforced configs.
   */
  public function save() {
    $this
      ->setConfigFilePath()
      ->prepareConfigDirectory()
      ->initializeNewConfigFile()
      ->cleanStaleConfigFile();

    $this->cleanStaleConfigMovedToNewTargetModule();

    (new EnforcedConfigRegistry($this->targetModule))->getEditableConfig()
      ->set(self::encode($this->configName), [
        'target_module' => $this->targetModule,
        'config_directory' => $this->configDirectory,
        'config_file_path' => $this->configFilePath,
        'enforcement_level' => $this->enforcementLevel,
        'config_form_uri' => $this->configFormUri,
        'hash' => $this->generateHash(),
      ])
      ->save();
    return $this;
  }

  /**
   * Clean up stale enforced config that is moving to another target module.
   */
  protected function cleanStaleConfigMovedToNewTargetModule() {
    if (!$this->configIsMovingToNewTargetModule()) return;
    (new EnforcedConfigRegistry($this->targetModule))->getEditableConfig()
      ->clear(self::encode($this->configName))
      ->save();
  }

  /**
   * Determine whether enforced config is being moved to another target module.
   */
  protected function configIsMovingToNewTargetModule() {
    return $this->getOldTargetModule() !== $this->targetModule;
  }

  /**
   * Return the current target module for the current config.
   */
  protected function getOldTargetModule() {
    return self::getTargetModule($this->configName);
  }

  /**
   * Delete an existing config file, if its path has changed.
   */
  protected function cleanStaleConfigFile() {
    if ($this->configFilePathChanged()) {
      \Drupal::service('file_system')
        ->delete(self::getConfigFilePath($this->configName));
    }
    return $this;
  }

  /**
   * Create a new config file.
   *
   * Config Devel is generally responsible for writing enforced config to disk.
   * However, when we enable a config for the first time, it won't yet be in
   * the list of enforced configs the Config Devel should write to disk. So, we
   * have to write it ourselves this time.
   *
   * @see EnforcedConfig::autoExport()
   */
  protected function initializeNewConfigFile() {
    if (!$this->configFileExists()) {
      \Drupal::service('config_devel.writeback_subscriber')
        ->writeBackConfig(\Drupal::config($this->configName), [$this->configFilePath]);
    }
    return $this;
  }

  /**
   * Determine whether the path for a config file has changed.
   */
  protected function configFilePathChanged() {
    // If the config file doesn't exist yet, then it hasn't changed.
    if (!$this->configFileExists()) return FALSE;

    return $this->configFilePath !== self::getConfigFilePath($this->configName);
  }

  /**
   * Determine whether a config file exists.
   */
  protected function configFileExists() {
    return file_exists(self::getConfigFilePath($this->configName));
  }

  /**
   * Ensure that the directory, in which a config file should be written, exists.
   */
  protected function prepareConfigDirectory() {
    $file_system = \Drupal::service('file_system');
    $directory = $file_system->dirname($this->configFilePath);
    $options = FileSystemInterface::CREATE_DIRECTORY + FileSystemInterface::MODIFY_PERMISSIONS;
    $file_system->prepareDirectory($directory, $options);
    return $this;
  }

  /**
   * Derive the config file path from the target module and config directory.
   */
  protected function setConfigFilePath() {
    $this->configFilePath = drupal_get_path('module', $this->targetModule);
    $this->configFilePath .= '/' . $this->configDirectory . '/';
    $this->configFilePath .= $this->configName . '.yml';
    return $this;
  }

}
